@@ -1,6 +1,9 @@
 module.exports = exports = configure
-module.exports.test = { findAccessibleSync: findAccessibleSync, 
-  findPython: findPython }
+module.exports.test = {
+  PythonFinder: PythonFinder,
+  findAccessibleSync: findAccessibleSync,
+  findPython: findPython,
+}
 
 /**
  * Module dependencies.
@@ -14,14 +17,13 @@ var fs = require('graceful-fs')
   , semver = require('semver')
   , mkdirp = require('mkdirp')
   , cp = require('child_process')
-  , PathArray = require('path-array')
   , extend = require('util')._extend
   , processRelease = require('./process-release')
-  , spawn = cp.spawn
-  , execFile = cp.execFile
-  , win = process.platform == 'win32'
+  , win = process.platform === 'win32'
   , findNodeDirectory = require('./find-node-directory')
   , msgFormat = require('util').format
+if (win)
+  var findVS2017 = require('./find-vs2017')
 
 exports.usage = 'Generates ' + (win ? 'MSVC project files' : 'a Makefile') + ' for the current module'
 
@@ -86,11 +88,22 @@ function configure (gyp, argv, callback) {
     mkdirp(buildDir, function (err, isNew) {
       if (err) return callback(err)
       log.verbose('build dir', '"build" dir needed to be created?', isNew)
-      createConfigFile()
+      if (win && (!gyp.opts.msvs_version || gyp.opts.msvs_version === '2017')) {
+        findVS2017(function (err, vsSetup) {
+          if (err) {
+            log.verbose('Not using VS2017:', err.message)
+            createConfigFile()
+          } else {
+            createConfigFile(null, vsSetup)
+          }
+        })
+      } else {
+        createConfigFile()
+      }
     })
   }
 
-  function createConfigFile (err) {
+  function createConfigFile (err, vsSetup) {
     if (err) return callback(err)
 
     var configFilename = 'config.gypi'
@@ -131,12 +144,22 @@ function configure (gyp, argv, callback) {
     // set the node development directory
     variables.nodedir = nodeDir
 
-    // don't copy dev libraries with nodedir option
-    variables.copy_dev_lib = !gyp.opts.nodedir
-
     // disable -T "thin" static archives by default
     variables.standalone_static_library = gyp.opts.thin ? 0 : 1
 
+    if (vsSetup) {
+      // GYP doesn't (yet) have support for VS2017, so we force it to VS2015
+      // to avoid pulling a floating patch that has not landed upstream.
+      // Ref: https://chromium-review.googlesource.com/#/c/433540/
+      gyp.opts.msvs_version = '2015'
+      process.env['GYP_MSVS_VERSION'] = 2015
+      process.env['GYP_MSVS_OVERRIDE_PATH'] = vsSetup.path
+      defaults['msbuild_toolset'] = 'v141'
+      defaults['msvs_windows_target_platform_version'] = vsSetup.sdk
+      variables['msbuild_path'] = path.join(vsSetup.path, 'MSBuild', '15.0',
+                                            'Bin', 'MSBuild.exe')
+    }
+
     // loop through the rest of the opts and add the unknown ones as variables.
     // this allows for module-specific configure flags like:
     //
@@ -260,6 +283,9 @@ function configure (gyp, argv, callback) {
         output_dir = buildDir
       }
       var nodeGypDir = path.resolve(__dirname, '..')
+      var nodeLibFile = path.join(nodeDir,
+        !gyp.opts.nodedir ? '<(target_arch)' : '$(Configuration)',
+        release.name + '.lib')
 
       argv.push('-I', addon_gypi)
       argv.push('-I', common_gypi)
@@ -270,8 +296,10 @@ function configure (gyp, argv, callback) {
         argv.push('-Dnode_exp_file=' + node_exp_file)
       }
       argv.push('-Dnode_gyp_dir=' + nodeGypDir)
-      argv.push('-Dnode_lib_file=' + release.name + '.lib')
+      argv.push('-Dnode_lib_file=' + nodeLibFile)
       argv.push('-Dmodule_root_dir=' + process.cwd())
+      argv.push('-Dnode_engine=' +
+        (gyp.opts.node_engine || process.jsEngine || 'v8'))
       argv.push('--depth=.')
       argv.push('--no-parallel')
 
@@ -288,8 +316,11 @@ function configure (gyp, argv, callback) {
       argv.unshift(gyp_script)
 
       // make sure python uses files that came with this particular node package
-      var pypath = new PathArray(process.env, 'PYTHONPATH')
-      pypath.unshift(path.join(__dirname, '..', 'gyp', 'pylib'))
+      var pypath = [path.join(__dirname, '..', 'gyp', 'pylib')]
+      if (process.env.PYTHONPATH) {
+        pypath.push(process.env.PYTHONPATH)
+      }
+      process.env.PYTHONPATH = pypath.join(win ? ';' : ':')
 
       var cp = gyp.spawn(python, argv)
       cp.on('exit', onCpExit)
@@ -312,9 +343,9 @@ function configure (gyp, argv, callback) {
 }
 
 /**
- * Returns the first file or directory from an array of candidates that is 
+ * Returns the first file or directory from an array of candidates that is
  * readable by the current user, or undefined if none of the candidates are
- * readable. 
+ * readable.
  */
 function findAccessibleSync (logprefix, dir, candidates) {
   for (var next = 0; next < candidates.length; next++) {
@@ -334,34 +365,47 @@ function findAccessibleSync (logprefix, dir, candidates) {
   return undefined
 }
 
-function findPython (python, callback) {
-  checkPython()
+function PythonFinder(python, callback) {
+  this.callback = callback
+  this.python = python
+}
 
-  // Check if Python is in the $PATH
-  function checkPython () {
-    log.verbose('check python', 'checking for Python executable "%s" in the PATH', python)
-    which(python, function (err, execPath) {
+PythonFinder.prototype = {
+  checkPythonLauncherDepth: 0,
+  env: process.env,
+  execFile: cp.execFile,
+  log: log,
+  resolve: path.win32 && path.win32.resolve || path.resolve,
+  stat: fs.stat,
+  which: which,
+  win: win,
+
+  checkPython: function checkPython () {
+    this.log.verbose('check python',
+                     'checking for Python executable "%s" in the PATH',
+                     this.python)
+    this.which(this.python, function (err, execPath) {
       if (err) {
-        log.verbose('`which` failed', python, err)
-        if (python === 'python2') {
-          python = 'python'
-          return checkPython()
+        this.log.verbose('`which` failed', this.python, err)
+        if (this.python === 'python2') {
+          this.python = 'python'
+          return this.checkPython()
         }
-        if (win) {
-          checkPythonLauncher()
+        if (this.win) {
+          this.checkPythonLauncher()
         } else {
-          failNoPython()
+          this.failNoPython()
         }
       } else {
-        log.verbose('`which` succeeded', python, execPath)
-        // Found the `python` exceutable, and from now on we use it explicitly.
+        this.log.verbose('`which` succeeded', this.python, execPath)
+        // Found the `python` executable, and from now on we use it explicitly.
         // This solves #667 and #750 (`execFile` won't run batch files
         // (*.cmd, and *.bat))
-        python = execPath
-        checkPythonVersion()
+        this.python = execPath
+        this.checkPythonVersion()
       }
-    })
-  }
+    }.bind(this))
+  },
 
   // Distributions of Python on Windows by default install with the "py.exe"
   // Python launcher which is more likely to exist than the Python executable
@@ -371,63 +415,48 @@ function findPython (python, callback) {
   // the first command line argument. Since "py.exe -2" would be an invalid
   // executable for "execFile", we have to use the launcher to figure out
   // where the actual "python.exe" executable is located.
-  function checkPythonLauncher () {
-    log.verbose('could not find "' + python + '". checking python launcher')
-    var env = extend({}, process.env)
+  checkPythonLauncher: function checkPythonLauncher () {
+    this.checkPythonLauncherDepth += 1
+
+    this.log.verbose(
+        'could not find "' + this.python + '". checking python launcher')
+    var env = extend({}, this.env)
     env.TERM = 'dumb'
 
     var launcherArgs = ['-2', '-c', 'import sys; print sys.executable']
-    execFile('py.exe', launcherArgs, { env: env }, function (err, stdout) {
+    this.execFile('py.exe', launcherArgs, { env: env }, function (err, stdout) {
       if (err) {
-        guessPython()
-        return
+        this.guessPython()
+      } else {
+        this.python = stdout.trim()
+        this.log.verbose('check python launcher',
+                         'python executable found: %j',
+                         this.python)
+        this.checkPythonVersion()
       }
-      python = stdout.trim()
-      log.verbose('check python launcher', 'python executable found: %j', python)
-      checkPythonVersion()
-    })
-  }
+      this.checkPythonLauncherDepth -= 1
+    }.bind(this))
+  },
 
-  // Called on Windows when "python" isn't available in the current $PATH.
-  // We're gonna check if "%SystemDrive%\python27\python.exe" exists.
-  function guessPython () {
-    log.verbose('could not find "' + python + '". guessing location')
-    var rootDir = process.env.SystemDrive || 'C:\\'
-    if (rootDir[rootDir.length - 1] !== '\\') {
-      rootDir += '\\'
-    }
-    var pythonPath = path.resolve(rootDir, 'Python27', 'python.exe')
-    log.verbose('ensuring that file exists:', pythonPath)
-    fs.stat(pythonPath, function (err, stat) {
-      if (err) {
-        if (err.code == 'ENOENT') {
-          failNoPython()
-        } else {
-          callback(err)
-        }
-        return
-      }
-      python = pythonPath
-      checkPythonVersion()
-    })
-  }
-
-  function checkPythonVersion () {
-    var env = extend({}, process.env)
+  checkPythonVersion: function checkPythonVersion () {
+    var args = ['-c', 'import platform; print(platform.python_version());']
+    var env = extend({}, this.env)
     env.TERM = 'dumb'
 
-    execFile(python, ['-c', 'import platform; print(platform.python_version());'], { env: env }, function (err, stdout) {
+    this.execFile(this.python, args, { env: env }, function (err, stdout) {
       if (err) {
-        return callback(err)
+        return this.callback(err)
       }
-      log.verbose('check python version', '`%s -c "import platform; print(platform.python_version());"` returned: %j', python, stdout)
+      this.log.verbose('check python version',
+                       '`%s -c "' + args[1] + '"` returned: %j',
+                       this.python, stdout)
       var version = stdout.trim()
       if (~version.indexOf('+')) {
-        log.silly('stripping "+" sign(s) from version')
+        this.log.silly('stripping "+" sign(s) from version')
         version = version.replace(/\+/g, '')
       }
       if (~version.indexOf('rc')) {
-        log.silly('stripping "rc" identifier from version')
+        this.log.silly('stripping "rc" identifier from version')
         version = version.replace(/rc(.*)$/ig, '')
       }
       var range = semver.Range('>=2.5.0 <3.0.0')
@@ -435,24 +464,60 @@ function findPython (python, callback) {
       try {
         valid = range.test(version)
       } catch (e) {
-        log.silly('range.test() error', e)
+        this.log.silly('range.test() error', e)
       }
       if (valid) {
-        callback(null, python)
+        this.callback(null, this.python)
+      } else if (this.win && this.checkPythonLauncherDepth === 0) {
+        this.checkPythonLauncher()
       } else {
-        failPythonVersion(version)
+        this.failPythonVersion(version)
       }
-    })
-  }
+    }.bind(this))
+  },
 
-  function failNoPython () {
-    callback(new Error('Can\'t find Python executable "' + python +
-          '", you can set the PYTHON env variable.'))
-  }
+  failNoPython: function failNoPython () {
+    var errmsg =
+        'Can\'t find Python executable "' + this.python +
+        '", you can set the PYTHON env variable.'
+    this.callback(new Error(errmsg))
+  },
 
-  function failPythonVersion (badVersion) {
-    callback(new Error('Python executable "' + python +
-          '" is v' + badVersion + ', which is not supported by gyp.\n' +
-          'You can pass the --python switch to point to Python >= v2.5.0 & < 3.0.0.'))
-  }
+  failPythonVersion: function failPythonVersion (badVersion) {
+    var errmsg =
+        'Python executable "' + this.python +
+        '" is v' + badVersion + ', which is not supported by gyp.\n' +
+        'You can pass the --python switch to point to ' +
+        'Python >= v2.5.0 & < 3.0.0.'
+    this.callback(new Error(errmsg))
+  },
+
+  // Called on Windows when "python" isn't available in the current $PATH.
+  // We are going to check if "%SystemDrive%\python27\python.exe" exists.
+  guessPython: function guessPython () {
+    this.log.verbose('could not find "' + this.python + '". guessing location')
+    var rootDir = this.env.SystemDrive || 'C:\\'
+    if (rootDir[rootDir.length - 1] !== '\\') {
+      rootDir += '\\'
+    }
+    var pythonPath = this.resolve(rootDir, 'Python27', 'python.exe')
+    this.log.verbose('ensuring that file exists:', pythonPath)
+    this.stat(pythonPath, function (err, stat) {
+      if (err) {
+        if (err.code == 'ENOENT') {
+          this.failNoPython()
+        } else {
+          this.callback(err)
+        }
+        return
+      }
+      this.python = pythonPath
+      this.checkPythonVersion()
+    }.bind(this))
+  },
+}
+
+function findPython (python, callback) {
+  var finder = new PythonFinder(python, callback)
+  finder.checkPython()
 }
